==> FailSafeC-1.0.1.1735.tar.gz: FailSafeC-1.0.1.1735/fscw/wl.ml <==
open Ctt_abstree
open Translate

type wl_struct = {
  (* struct tag name *)
  wl_struct_tag: string option;
  (* __fsc_attribute__ name *)
  wl_struct_name: string option;
  wl_struct_id: struct_id;
}

(*
 * value of wrapped type
 *  fscw_void, fscw_int, fscw_pointer, ...
 *)
type wl_type =
| WLTvoid
| WLTbuiltin of builtin_type
| WLTpointer
| WLTstruct of wl_struct
| WLTabstract of string
| WLTstring

(*
 * global names.
 *)
type wl_global_decl = 
| WLTypedef of identifier * c_type          (* typedef c_type identifier; *)
| WLDeclVar of identifier * c_type * c_type (* extern  c_type identifier; *)
| WLStruct  of wl_struct                    (* struct tag { ... }; *)

type environment2 = {
  genv: Ctt_abstree.environment;
  decls: wl_global_decl list;
}

let make_genv2 ge dec = { genv = ge; decls = dec }

let c_type_of_typename ~genv2 name =
  let rec f x = match x with
    | [] -> failwith ("type name not found: " ^ name)
    | WLTypedef(n, ty)::_ when name = n -> ty
    | _::tl -> f tl in
  f genv2.decls

let c_type_of_varname ~genv2 name =
  let rec f x = match x with
    | [] -> failwith ("variable name not found: " ^ name)
    | WLDeclVar(n, abs_ty, ty)::_ when name = n -> abs_ty, ty
    | _::tl -> f tl in
  f genv2.decls

let wl_struct_of_struct_tag ~genv2 name =
  let rec f x = match x with
    | [] -> failwith ("struct tag not found: " ^ name)
    | WLStruct(s)::_ when s.wl_struct_tag = Some name -> s
    | _::tl -> f tl in
  f genv2.decls

let wl_struct_of_struct_name ~genv2 name =
  let rec f x = match x with
    | [] -> failwith ("struct name not found: " ^ name)
    | WLStruct(s)::_ when s.wl_struct_name = Some name -> s
    | _::tl -> f tl in
  f genv2.decls

let wl_struct_of_struct_id ~genv2 id =
  let rec f x = match x with
    | [] -> raise Not_found
    | WLStruct(s)::_ when s.wl_struct_id = id -> s
    | _::tl -> f tl in
  f genv2.decls

let wl_type_of_c_type ~genv2 ty =
  match ty.ct_ty with
  | Tvoid -> WLTvoid
  | Tbuiltin bt -> WLTbuiltin bt
  | Tpointer base -> begin
    match base.ct_ty with
    | Tbuiltin Tchar -> WLTstring (* TODO: need annotation syntax *)
    | _ -> WLTpointer 
    end
  | Tfunction _ -> assert false (* TODO *)
  | Tarray _ -> assert false (* TODO *)
  | Tstruct id -> WLTstruct (wl_struct_of_struct_id ~genv2 id)
  | Tabstract name -> WLTabstract name

let wl_type_of_param_c_type ~genv2 ty =
  match ty.ct_ty with
  | Tvoid -> WLTvoid
  | Tbuiltin bt -> WLTbuiltin bt
  | Tpointer base -> begin
    match base.ct_ty with
    | Tbuiltin Tchar -> WLTstring (* TODO: need annotation syntax *)
    | _ -> WLTpointer 
    end
  | Tfunction _ -> WLTpointer
  | Tarray _ -> WLTpointer
  | Tstruct id -> WLTstruct (wl_struct_of_struct_id ~genv2 id)
  | Tabstract name -> WLTabstract name

let wl_type_of_typename ~genv2 name =
  wl_type_of_c_type ~genv2 (c_type_of_typename ~genv2 name)

let canonify_c_type_mem t =
  Il0_type_canonify.translate_c_type_mem t

type wl_type_property = {
  tp_wl_type: wl_type;
  tp_fscw_t: c_type; (* Tabstract "fscw_T" *)
  tp_c_t: c_type;    (* T *)
  tp_c_t_str: string; (* str of T *)
  tp_param_is_native: bool;  (* param is native value (i.e. all builtin types) *)
  tp_param_has_unwrap_param: bool; (* fscw_T_unwrap_param is available (i.e. string) *)
  tp_prop: type_properties;
}

let rec parse_wl_type ~genv2 wty =
  let genv = genv2.genv in
  match wty with
  | WLTvoid -> assert false
  | WLTbuiltin bt -> begin
    let name, str_ty, bt' = match bt with
    | Tchar ->     "fscw_char", "char", Tuchar
    | Tschar ->    "fscw_signed_char", "signed char", Tuchar
    | Tuchar ->    "fscw_unsigned_char", "unsigned char", Tuchar
    | Tshort ->    "fscw_short", "short", Tushort
    | Tushort ->   "fscw_unsigned_short", "unsigned short", Tushort
    | Tint ->      "fscw_int", "int", Tuint
    | Tuint ->     "fscw_unsigned_int", "unsigned int", Tuint
    | Tlong ->     "fscw_long", "long", Tulong
    | Tulong ->    "fscw_unsigned_long", "unsigned long", Tulong
    | Tlonglong -> "fscw_long_long", "long long", Tulonglong
    | Tulonglong -> "fscw_unsigned_long_long", "unsigned long long", Tulonglong
    | Tfloat -> "fscw_float", "float", Tfloat
    | Tdouble -> "fscw_double", "double", Tdouble
    | Tlongdouble -> "fscw_long_double", "long double", Tlongdouble
    in
    let ty = make_c_type (Tbuiltin bt') in
    {
      tp_wl_type = wty;
      tp_fscw_t = make_c_type (Tabstract name);
      tp_c_t = make_c_type (Tbuiltin bt);
      tp_c_t_str = str_ty;
      tp_param_is_native = true; (* (short x) or (base_t x_base, int x) *)
      tp_param_has_unwrap_param = false;
      tp_prop = parse_type_genv ~genv ty;
    }
    end
  | WLTpointer ->
    let ty = make_c_type (Tpointer (make_c_type Tvoid)) in
    {
      tp_wl_type = wty;
      tp_fscw_t = make_c_type (Tabstract "fscw_pointer");
      tp_c_t = ty;
      tp_c_t_str = "void *";
      tp_param_is_native = false;
      tp_param_has_unwrap_param = false;
      tp_prop = parse_type_genv ~genv ty;
    }
  | WLTstruct s ->
    let n = match s.wl_struct_tag, s.wl_struct_name with
            | Some t, _ -> t
            | None, Some t -> t
            | None, None -> assert false in
    let ty = make_c_type ~s_table:genv.struct_table (Tstruct s.wl_struct_id) in
    {
      tp_wl_type = wty;
      tp_fscw_t = make_c_type (Tabstract ("fscw_struct_" ^ n));
      tp_c_t = ty;
      tp_c_t_str = "";
      tp_param_is_native = false;
      tp_param_has_unwrap_param = false;
      tp_prop = parse_type_genv ~genv ty;
    }
  | WLTabstract name ->
    let ty = c_type_of_typename ~genv2 name in
    let wty' = wl_type_of_c_type ~genv2 ty in (* resolve typedef name *)
    let p' = parse_wl_type ~genv2 wty' in
    {
      tp_wl_type = wty;
      tp_fscw_t = make_c_type (Tabstract ("fscw_" ^ name));
      tp_c_t = ty;
      tp_c_t_str = name;
      tp_param_is_native = p'.tp_param_is_native;
      tp_param_has_unwrap_param = p'.tp_param_has_unwrap_param;
      tp_prop = p'.tp_prop;
    }
  | WLTstring ->
    let ty = make_c_type (Tpointer (make_c_type (Tbuiltin Tchar))) in
    {
      tp_wl_type = wty;
      tp_fscw_t = make_c_type (Tabstract "fscw_string");
      tp_c_t = ty;
      tp_c_t_str = "const char *";
      tp_param_is_native = false;
      tp_param_has_unwrap_param = true;
      tp_prop = parse_type_genv ~genv ty;
    }

let fscw_T_of_prop p =
  match p.tp_fscw_t.ct_ty with
  | Tabstract name -> name
  | _ -> assert false

let fscw_T_of_wty ~genv2 wty =
  match wty with
  | WLTvoid -> "void"
  | _ -> fscw_T_of_prop (parse_wl_type ~genv2 wty)


(*

(*
 * "fscw_T" of c_type T
 *
 *
 *)
(*let typename_of_builtin_type bt = *)
let fscw_T_of_builtin_type bt =

let tagname_of_struct ~genv2 id =
  "struct_" ^ Translate.encoded_name_of_struct ~genv id

let typename_of_struct ~genv2 id =
  "struct " ^ tagname_of_struct ~genv id


let fscw_T_of_c_type ~genv2 tp =
  match tp.ct_ty with
      Tvoid -> "void"
    | Tbuiltin bt -> typename_of_builtin_type bt
    | Tpointer base_tp -> "fscw_pointer"
    | Tfunction _ -> "fscw_not_supported"
    | Tarray (base_tp, sz) ->  "fscw_not_supported"
    | Tstruct id ->  typename_of_struct ~genv id
    | Tabstract name ->  "fscw_" ^ name


let typeinfo_name_of_struct ~genv id =
  "fsc_typeinfo_" ^ Translate.encoded_name_of_struct ~genv id

let type_canonify = Il0_type_canonify.translate_c_type_mem

let typename_of_param_type ~genv tp =
  match tp.ct_ty with
      Tvoid -> "void"
    | Tbuiltin bt -> typename_of_builtin_type bt
    | Tpointer base_tp -> "fscw_pointer"
    | Tfunction _ -> "fscw_pointer"
    | Tarray (base_tp, sz) ->  "fscw_pointer"
    | Tstruct id ->  typename_of_struct ~genv id
    | Tabstract name ->  "fscw_" ^ name


let generate_header chan ~genv t gv =
    let printf x = Printf.fprintf chan x in
    let print_string = output_string chan in
    List.iter (function
       WLTypedef(name, ty, pty) ->
         printf "/* typedef %s = %s = %s */\n" name (Ctt_printer.strof_c_type ty) (Ctt_printer.strof_c_type pty);
        let old_ty = typename_of_c_type ~genv pty in
        let new_ty = "fscw_" ^ name in
        let tmpl =
          "#if !defined(FSCW_<newtp>_is_defined) && defined(FSCW_NEED_<newtp>)\n" ^
          "#define FSCW_<newtp>_is_defined\n" ^
          "typedef <oldtp> <newtp>;\n" ^
          "FSCW_INLINE <newtp> <newtp>_wrap(<tp> v){ return <oldtp>_wrap(v); }\n" ^
          "FSCW_INLINE <tp> <newtp>_unwrap(<newtp> v){ return <oldtp>_unwrap(v); }\n" ^
          "#define <newtp>_param(n) <oldtp>_param(n)\n" ^
          "#define <newtp>_member(t, m) <oldtp>_member(t, m)\n" ^
          "FSCW_INLINE <newtp> <newtp>_read(<newtp> *p){ return <oldtp>_read(p); }\n" ^
          "FSCW_INLINE void <newtp>_write(<newtp> *p, <newtp> v){ <oldtp>_write(p, v); }\n" ^
          "FSCW_INLINE <tp> <newtp>_read_unwrap(<newtp> *p){ return <oldtp>_read_unwrap(p); }\n" ^
          "FSCW_INLINE void <newtp>_wrap_write(<newtp> *p, <tp> v){ <oldtp>_wrap_write(p, v); }\n" ^
          "#define <newtp>_read_pointer_param(p, n) <oldtp>_read_pointer_param(p, n)\n" ^
          "#define <newtp>_write_pointer_param(p, n, v) <oldtp>_write_pointer_param(p, n, v)\n" ^
          "#endif\n\n"
        in
        let result = Str.global_substitute (Str.regexp "<[^>]+>")
           (fun str ->
              let s = Str.matched_string str in
              if s = "<newtp>" then new_ty
              else if s = "<oldtp>" then old_ty
              else name) tmpl
        in
        print_string result
     | WLDeclVar(name, ty, pty) -> begin
         printf "/* var     %s : %s = %s */\n" name (Ctt_printer.strof_c_type ty) (Ctt_printer.strof_c_type pty);
         let new_name = "fscw_" ^ name in
         match pty.ct_ty with
           Tfunction _ -> begin
           let old_name = Translate.translate_global_function_name ~genv name (type_canonify pty) in
           printf "#if !defined(FSCW_%s_is_declared) && defined(FSCW_NEED_%s)\n" name name;
           printf "#define FSCW_%s_is_declared\n" name;
           printf "#define %s %s\n" new_name old_name;
           match ty.ct_ty with
             Tfunction (param, is_var, ret_ty) ->
             printf "extern %s %s(" (typename_of_c_type ~genv ret_ty) new_name;
             let i = ref 1 in
             let sep = ref "" in
             List.iter
               (fun pt ->
                 let pt_name = typename_of_param_type ~genv pt in
                 printf "%s%s_param(param_%d)" !sep pt_name !i;
                 i := !i + 1;
                 sep := ", "
               ) param;
             if param = [] && not is_var then
               printf "void"
             else if is_var then
               printf "%s%s_param(param_%d)" !sep "fscw_pointer" !i;
             printf ");\n"
           | _ -> ()
           end;
           printf "#endif\n\n"
         | _ -> ()
         end
     | WLStruct(name, id)  ->
         printf "/* struct  %s : %d*/\n" name id;
         let new_name = "fscw_struct_" ^ name in
         let old_name = typename_of_struct ~genv id in
         let tag_name = tagname_of_struct ~genv id in
         let ti = typeinfo_name_of_struct ~genv id in
         let sz = Option.map Big_int.string_of_big_int (Ctt_abstree.get_struct_desc ~genv id).str_size in
         let tmpl = 
           "#if !defined(FSCW_<newtp>_is_defined) && defined(FSCW_NEED_<newtp>)\n" ^
           "#define FSCW_<newtp>_is_defined\n" ^
           (* Buffer.contents buf ^ *)
           "typedef <oldtp> <newtp>;\n" ^
           "extern <newtp> read_<tag>(base_t, ofs_t);\n" ^
           "extern void write_<tag>(base_t, ofs_t, <newtp>);\n" ^
           (match sz with
             Some _ ->
             "#define <newtp>_read_pointer_param(p, n) (read_<tag>(p##_base, p##_ofs + <sz> * n))\n" ^
             "#define <newtp>_write_pointer_param(p, n, v) (write_<tag>(p##_base, p##_ofs + <sz> * n, v))\n"
            | _ -> "") ^
           "extern struct typeinfo_init <ti>;\n" ^
           "FSCW_INLINE <newtp> *<newtp>_alloc(size_t n) {\n" ^
           "  return (<newtp> *)fsc_alloc_static_block(&<ti>.val, n);\n" ^
           "}\n" ^
           "#endif\n\n"
         in
         let result = Str.global_substitute (Str.regexp "<[^>]+>")
           (fun str ->
              let s = Str.matched_string str in
              if s = "<newtp>" then new_name
              else if s = "<oldtp>" then old_name
              else if s = "<ti>" then ti
              else if s = "<sz>" then Option.get sz
              else tag_name) tmpl in
          print_string result
      ) gv

let is_simple_type tp =
  match tp.ct_ty with
      Tvoid -> true
    | Tbuiltin bt -> true
    | Tpointer base_tp -> false
    | Tfunction _ -> false
    | Tarray (base_tp, sz) -> false
    | Tstruct id -> false
    | Tabstract name -> false

let strof_function_type ~genv ty name =
  let buf = Buffer.create 64 in
  let ppf = Format.formatter_of_buffer buf in
  let (specs, decl) = Ctt_to_ptree.convert_declarator ~genv ty name in
  Format.fprintf ppf "%a %a@?"
    C_pp.pp_print_sqlist specs
    C_pp.pp_print_declarator decl;
  Buffer.contents buf

let generate_impl chan ~genv t gv name =
  let printf x = Printf.fprintf chan x in
  let print_string = output_string chan in
  List.iter (function
      WLDeclVar(n, ty, pty) ->
      let cty = (type_canonify pty) in
      if n = name then begin
        match cty.ct_ty with
          Tfunction (param, is_var, ret_ty) ->
          if not (is_simple_type ret_ty) || not (List.for_all is_simple_type param) then
            failwith "not simple type";
          let ret_name = typename_of_c_type ~genv ret_ty in
          let fun_name = Translate.translate_global_function_name ~genv name cty in

          print_string "#define FSC_RUNTIME_LIBRARY\n";
          print_string "#include <fsc_runtime.h>\n";
          print_string "#include <wrapper_helper.h>\n";
          print_string "#include <wrapper/fscw.h>\n";
          print_string "\n";

          print_string "/* manifest entry\n";
          printf "D\t%s\t%s\n" name (Translate.encoded_name_of_type_genv ~genv cty);
          print_string "*/\n";
          print_string "\n";

          print_string (Ctt_printer.strof_global_declaration (CTTdeclVariable(Extern [], pty, name, None)));

          print_string "/**\n";
          print_string " * @fn ";
          print_string (strof_function_type ~genv ty name);
          print_string "\n";
          print_string " * @brief auto generated function.\n";
          print_string " * @author auto generated\n";
          print_string " */\n";

          printf "%s %s(%s)\n" ret_name fun_name
            (String.concat ", "
              (ExtList.List.mapi
                (fun i param_ty ->
                  Printf.sprintf "%s_param(_param_%d)" (typename_of_c_type ~genv param_ty) i
                ) param));
          print_string "{\n";
          let unwraps = 
            String.concat ", "
              (ExtList.List.mapi
                (fun i param_ty ->
                  Printf.sprintf "%s_unwrap(_param_%d)" (typename_of_c_type ~genv param_ty) i
                ) param) in
          if ret_ty.ct_ty = Tvoid then
            printf "  %s(%s);\n" name unwraps
          else
            printf "  return %s_wrap(%s(%s));\n" ret_name name unwraps;
          print_string "}\n"
        | _ -> ()
      end
    | _ -> ()) gv
*)

==> FailSafeC-1.5.0.1797.tar.gz: FailSafeC-1.5.0.1797/fscw/wl.ml <==
open Ctt_abstree
open Translate

type wl_struct = {
  (* struct tag name *)
  wl_struct_tag: string option;
  (* __fsc_attribute__ name *)
  wl_struct_name: string option;
  wl_struct_id: struct_id;
}

(*
 * value of wrapped type
 *  fscw_void, fscw_int, fscw_pointer, ...
 *)
type wl_type =
| WLTvoid
| WLTbuiltin of builtin_type
| WLTpointer
| WLTstruct of wl_struct
| WLTabstract of string
| WLTstring

(*
 * global names.
 *)
type wl_global_decl = 
| WLTypedef of identifier * c_type          (* typedef c_type identifier; *)
| WLDeclVar of identifier * c_type * c_type (* extern  c_type identifier; *)
| WLStruct  of wl_struct                    (* struct tag { ... }; *)

type environment2 = {
  genv: Ctt_abstree.environment;
  decls: wl_global_decl list;
  depends: (identifier * identifier list) list;
  original_decls: (identifier * C_abstree.declaration_desc) list
}

let make_genv2 ge dec dep orig = { genv = ge; decls = dec; depends = dep; original_decls = orig; }

let c_type_of_typename ~genv2 name =
  let rec f x = match x with
    | [] -> failwith ("type name not found: " ^ name)
    | WLTypedef(n, ty)::_ when name = n -> ty
    | _::tl -> f tl in
  f genv2.decls

let c_type_of_varname ~genv2 name =
  let rec f x = match x with
    | [] -> failwith ("variable name not found: " ^ name)
    | WLDeclVar(n, abs_ty, ty)::_ when name = n -> abs_ty, ty
    | _::tl -> f tl in
  f genv2.decls

let wl_struct_of_struct_tag ~genv2 name =
  let rec f x = match x with
    | [] -> failwith ("struct tag not found: " ^ name)
    | WLStruct(s)::_ when s.wl_struct_tag = Some name -> s
    | _::tl -> f tl in
  f genv2.decls

let wl_struct_of_struct_name ~genv2 name =
  let rec f x = match x with
    | [] -> failwith ("struct name not found: " ^ name)
    | WLStruct(s)::_ when s.wl_struct_name = Some name -> s
    | _::tl -> f tl in
  f genv2.decls

let wl_struct_of_struct_id ~genv2 id =
  let rec f x = match x with
    | [] -> raise Not_found
    | WLStruct(s)::_ when s.wl_struct_id = id -> s
    | _::tl -> f tl in
  f genv2.decls

let wl_type_of_c_type ~genv2 ty =
  match ty.ct_ty with
  | Tvoid -> WLTvoid
  | Tbuiltin bt -> WLTbuiltin bt
  | Tpointer base -> begin
    match base.ct_ty with
    | Tbuiltin Tchar -> WLTstring (* TODO: need annotation syntax *)
    | _ -> WLTpointer 
    end
  | Tfunction _ -> assert false (* TODO *)
  | Tarray _ -> assert false (* TODO *)
  | Tstruct id -> WLTstruct (wl_struct_of_struct_id ~genv2 id)
  | Tabstract name -> WLTabstract name

let wl_type_of_param_c_type ~genv2 ty =
  match ty.ct_ty with
  | Tvoid -> WLTvoid
  | Tbuiltin bt -> WLTbuiltin bt
  | Tpointer base -> begin
    match base.ct_ty with
    | Tbuiltin Tchar -> WLTstring (* TODO: need annotation syntax *)
    | _ -> WLTpointer 
    end
  | Tfunction _ -> WLTpointer
  | Tarray _ -> WLTpointer
  | Tstruct id -> WLTstruct (wl_struct_of_struct_id ~genv2 id)
  | Tabstract name -> WLTabstract name

let wl_type_of_typename ~genv2 name =
  wl_type_of_c_type ~genv2 (c_type_of_typename ~genv2 name)

let canonify_c_type_mem t =
  Il0_type_canonify.translate_c_type_mem t

type wl_type_property = {
  tp_wl_type: wl_type;
  tp_fscw_t: c_type; (* Tabstract "fscw_T" *)
  tp_c_t: c_type;    (* T *)
  tp_c_t_str: string; (* str of T *)
  tp_param_is_native: bool;  (* param is native value (i.e. all builtin types) *)
  tp_param_has_unwrap_param: bool; (* fscw_T_unwrap_param is available (i.e. string) *)
  tp_prop: type_properties;
}

let rec parse_wl_type ~genv2 wty =
  let genv = genv2.genv in
  match wty with
  | WLTvoid -> assert false
  | WLTbuiltin bt -> begin
    let name, str_ty, bt' = match bt with
    | Tchar ->     "fscw_char", "char", Tuchar
    | Tschar ->    "fscw_signed_char", "signed char", Tuchar
    | Tuchar ->    "fscw_unsigned_char", "unsigned char", Tuchar
    | Tshort ->    "fscw_short", "short", Tushort
    | Tushort ->   "fscw_unsigned_short", "unsigned short", Tushort
    | Tint ->      "fscw_int", "int", Tuint
    | Tuint ->     "fscw_unsigned_int", "unsigned int", Tuint
    | Tlong ->     "fscw_long", "long", Tulong
    | Tulong ->    "fscw_unsigned_long", "unsigned long", Tulong
    | Tlonglong -> "fscw_long_long", "long long", Tulonglong
    | Tulonglong -> "fscw_unsigned_long_long", "unsigned long long", Tulonglong
    | Tfloat -> "fscw_float", "float", Tfloat
    | Tdouble -> "fscw_double", "double", Tdouble
    | Tlongdouble -> "fscw_long_double", "long double", Tlongdouble
    in
    let ty = make_c_type (Tbuiltin bt') in
    {
      tp_wl_type = wty;
      tp_fscw_t = make_c_type (Tabstract name);
      tp_c_t = make_c_type (Tbuiltin bt);
      tp_c_t_str = str_ty;
      tp_param_is_native = true; (* (short x) or (base_t x_base, int x) *)
      tp_param_has_unwrap_param = false;
      tp_prop = parse_type_genv ~genv ty;
    }
    end
  | WLTpointer ->
    let ty = make_c_type (Tpointer (make_c_type Tvoid)) in
    {
      tp_wl_type = wty;
      tp_fscw_t = make_c_type (Tabstract "fscw_pointer");
      tp_c_t = ty;
      tp_c_t_str = "void *";
      tp_param_is_native = false;
      tp_param_has_unwrap_param = false;
      tp_prop = parse_type_genv ~genv ty;
    }
  | WLTstruct s ->
    let n = match s.wl_struct_tag, s.wl_struct_name with
            | _, Some t -> t
            | Some t, None -> t
            | None, None -> assert false in
    let ty = make_c_type ~s_table:genv.struct_table (Tstruct s.wl_struct_id) in
    {
      tp_wl_type = wty;
      tp_fscw_t = make_c_type (Tabstract ("fscw_struct_" ^ n));
      tp_c_t = ty;
      tp_c_t_str = "";
      tp_param_is_native = false;
      tp_param_has_unwrap_param = false;
      tp_prop = parse_type_genv ~genv ty;
    }
  | WLTabstract name ->
    let ty = c_type_of_typename ~genv2 name in
    let wty' = wl_type_of_c_type ~genv2 ty in (* resolve typedef name *)
    let p' = parse_wl_type ~genv2 wty' in
    {
      tp_wl_type = wty;
      tp_fscw_t = make_c_type (Tabstract ("fscw_" ^ name));
      tp_c_t = ty;
      tp_c_t_str = name;
      tp_param_is_native = p'.tp_param_is_native;
      tp_param_has_unwrap_param = p'.tp_param_has_unwrap_param;
      tp_prop = p'.tp_prop;
    }
  | WLTstring ->
    let ty = make_c_type (Tpointer (make_c_type (Tbuiltin Tchar))) in
    {
      tp_wl_type = wty;
      tp_fscw_t = make_c_type (Tabstract "fscw_string");
      tp_c_t = ty;
      tp_c_t_str = "const char *";
      tp_param_is_native = false;
      tp_param_has_unwrap_param = true;
      tp_prop = parse_type_genv ~genv ty;
    }

let fscw_T_of_prop p =
  match p.tp_fscw_t.ct_ty with
  | Tabstract name -> name
  | _ -> assert false

let fscw_T_of_wty ~genv2 wty =
  match wty with
  | WLTvoid -> "void"
  | _ -> fscw_T_of_prop (parse_wl_type ~genv2 wty)
